unit BitPulseStreamers;

interface

uses
  Windows, Classes;

type
  TBitPulseStreamer = class(TThread)
  protected
    procedure Execute; override; //Basically the threadproc.
  end;

procedure ClearBitPulses;
function SetBitPulse(BitIndex: LongWord; NewHighSide: LongWord; NewLowSide: LongWord): LongWord;
function BeginStreaming: LongWord;
function EndStreaming: LongWord;

var
  StreamDeviceIndex: LongWord;
  StreamErr: LongWord;



implementation

uses
  AIOUSB;

type
  TBitPulseData = record
    SideDuration: array[0..1] of LongWord;

    PlayingSide: Integer; //0 for low side, 1 for high side.
    PlayingTimeLeft: LongWord;
  end;

var
  BitPulseData: array[0..15] of TBitPulseData;
  hBPDMutex: THandle = INVALID_HANDLE_VALUE;
  Streamer: TBitPulseStreamer = nil;



procedure ClearBitPulses;
var
  iBit: Integer;

begin
  if hBPDMutex <> INVALID_HANDLE_VALUE then WaitForSingleObject(hBPDMutex, INFINITE);

  for iBit := 0 to 15 do begin
    BitPulseData[iBit].SideDuration[0] := 1;
    BitPulseData[iBit].SideDuration[1] := 0;
  end;

  if hBPDMutex <> INVALID_HANDLE_VALUE then ReleaseMutex(hBPDMutex);
end;

function SetBitPulse(BitIndex: LongWord; NewHighSide: LongWord; NewLowSide: LongWord): LongWord;
begin
  if BitIndex > 15 then begin
    Result := ERROR_INVALID_INDEX;
    Exit;
  end;

  if (NewHighSide = 0) and (NewLowSide = 0) then begin
    Result := ERROR_INVALID_PARAMETER;
    Exit;
  end;

  if hBPDMutex <> INVALID_HANDLE_VALUE then WaitForSingleObject(hBPDMutex, INFINITE);

  BitPulseData[BitIndex].SideDuration[0] := NewLowSide;
  BitPulseData[BitIndex].SideDuration[1] := NewHighSide;

  if hBPDMutex <> INVALID_HANDLE_VALUE then ReleaseMutex(hBPDMutex);

  Result := ERROR_SUCCESS;
end;

function BeginStreaming: LongWord;
var
  iBit: Integer;

begin
  if Streamer <> nil then begin
    Result := ERROR_ALREADY_ASSIGNED;
    Exit;
  end;

  for iBit := 0 to 15 do begin
    BitPulseData[iBit].PlayingSide := 0;
    BitPulseData[iBit].PlayingTimeLeft := 0;
  end;
  StreamErr := ERROR_SUCCESS;

  hBPDMutex := CreateMutex(nil, False, nil);

  Streamer := TBitPulseStreamer.Create(False);

  Result := ERROR_SUCCESS;
end;

function EndStreaming: LongWord;
begin
  if Streamer = nil then begin
    Result := ERROR_NOT_READY;
    Exit;
  end;

  Streamer.Terminate; //Sets the thread's Terminated.
  Streamer.Free; //Waits for the thread to finish, then deallocates it.
  Streamer := nil;

  CloseHandle(hBPDMutex);
  hBPDMutex := INVALID_HANDLE_VALUE;

  Result := ERROR_SUCCESS;
end;

{ TBitPulseStreamer }

procedure TBitPulseStreamer.Execute;
const
  BufWords = 32*1024; //32k word(64kB) data buffer.

var
  ReadHz, WriteHz: Double;
  iWord, iBit: Integer;
  ThisWord: Word;
  Buf: array[0..BufWords-1] of Word;
  BytesTransferred: LongWord;
  BitTerminated: Word;

begin
  ReadHz := 0; //Not reading.
  WriteHz := 1024*1024; //Writing at 1MHz.

  DIO_StreamOpen(StreamDeviceIndex, False);
  DIO_StreamSetClocks(StreamDeviceIndex, ReadHz, WriteHz);

  ThisWord := $0000;
  BitTerminated := $0000;
  while BitTerminated <> $FFFF do begin //Until all 16 bits have terminated.
    WaitForSingleObject(hBPDMutex, INFINITE);

    for iWord := 0 to BufWords-1 do begin
      //Build each word from the 16 mini state machines.
      for iBit := 0 to 15 do begin
        if BitPulseData[iBit].PlayingTimeLeft = 0 then begin //Time's up on the playing side.
          if Terminated then begin
            BitTerminated := BitTerminated or (1 shl iBit);
            //End each bit's waveform in its own time, settling back to low states. This continues "forever", i.e. until the stream finishes ending and gets restarted.
            BitPulseData[iBit].PlayingSide := 0;
            BitPulseData[iBit].PlayingTimeLeft := $FFFFFFFF;
          end
          else begin
            BitPulseData[iBit].PlayingSide := BitPulseData[iBit].PlayingSide xor 1; //Switch to the other side.
            BitPulseData[iBit].PlayingTimeLeft := BitPulseData[iBit].SideDuration[BitPulseData[iBit].PlayingSide]; //However many 1s ticks the new side plays for.

            if BitPulseData[iBit].PlayingTimeLeft = 0 then begin //The new side has no duration, so switch back.
              BitPulseData[iBit].PlayingSide := BitPulseData[iBit].PlayingSide xor 1;
              //Could optimize this by setting PlayingTimeLeft to however long is left in this buffer, but that interacts with BitTerminated, so it's kinda tricky.

              if BitPulseData[iBit].PlayingSide <> 0 then //Update that bit in the word.
                ThisWord := ThisWord or (1 shl iBit)
              else
                ThisWord := ThisWord and not (1 shl iBit)
              ;
            end;
          end;
        end;

        Dec(BitPulseData[iBit].PlayingTimeLeft);
      end;
      Buf[iWord] := ThisWord;
    end;

    ReleaseMutex(hBPDMutex);

    StreamErr := DIO_StreamFrame(StreamDeviceIndex, BufWords, @Buf[0], BytesTransferred);
    if StreamErr <> ERROR_SUCCESS then Break;
  end;
  DIO_StreamClose(StreamDeviceIndex);
end;

end.
